#!/usr/bin/env python2.6
# Copyright 2009-2010 David D Lowe
# Released under the GPL v3
# some code inspired by jockey (https://launchpad.net/~jockey)
"""This is the DBus service that provides the functionality for 
PolicyKitShell in shell.py"""
import dbus
import dbus.service
import dbus.mainloop
import dbus.glib
import gobject
import sys, os
import subprocess
import const
from const import logger

class UnauthorisedPID(dbus.DBusException):
    _dbus_error_name = "org.tuxfamily.epidermis.Shell.UnauthorisedPID"

class __PShell(dbus.service.Object):
    def __init__(self, mainloop, conn=None, object_path=None, bus_name=None):
        dbus.service.Object.__init__(self,conn,object_path,bus_name)
        
        self.mainloop = mainloop
        
        self.authorisedPids = {}
        
        # the following variables are used by _check_polkit_privilege
        self.dbus_info = None
        self.polkit = None
        self.enforce_polkit = True
    
    @dbus.service.method("org.tuxfamily.epidermis.PShellInterface",
                         in_signature='', out_signature='b',
                         sender_keyword='sender', connection_keyword='conn')
    def prepare(self, sender=None, conn=None):
        """Prepares this service to accept do() commands from the process
        that is calling this method, identified by its PID and start-time.
        This should ask the user for a password thanks to PolicyKit."""
        self.__setup_dbus_info(conn)
        pid = self.dbus_info.GetConnectionUnixProcessID(sender)
        start_time = self.__get_start_time(pid)
        try:
            self.__verify_pid_is_epidermis(pid)
        except UnauthorisedPID, ee:
            self.__verify_pids()
            raise ee
        
        if self.polkit is None:
            self.polkit = dbus.Interface(dbus.SystemBus().get_object(
                'org.freedesktop.PolicyKit1',
                '/org/freedesktop/PolicyKit1/Authority', False),
                'org.freedesktop.PolicyKit1.Authority')
        
        try:
            (is_auth, _, details) = self.polkit.CheckAuthorization(
                ('unix-process', 
                     {'pid': dbus.UInt32(pid, variant_level=1),
                     'start-time':dbus.UInt64(start_time,variant_level=1)}), 
                "org.tuxfamily.epidermis.shell.managepigments",
                {'': ''}, dbus.UInt32(1), '', timeout=600)
            if is_auth:
                # add pid to dictionary of authorised PIDs
                self.authorisedPids[pid] = self.__get_start_time(pid)
            self.__verify_pids()
            return is_auth
        except dbus.DBusException, e:
            if e._dbus_error_name == 'org.freedesktop.DBus.Error.ServiceUnknown':
                # polkitd timed out, connect again
                self.polkit = None
                return self.prepare(sender=sender, conn=conn)
            else:
                raise
    
    @dbus.service.method("org.tuxfamily.epidermis.PShellInterface",
                        in_signature='asbs', out_signature='(iss)',
                        sender_keyword='sender', connection_keyword='conn')
    def do(self, command, intrShell=False, cwd=None, sender=None, conn=None):
        """Perform a command as root.
        command: a list of strings of the command to be run as root
        intrShell: if True, allows special characters such as * to be 
        interpreted by the shell
        cwd: current working directory (string)
        Returns (exitcode, output, error_output)
        Before using do, a process must call prepare() once. do() does not
        ask for PolicyKit authentication once prepare() has been called."""
        self.__setup_dbus_info(conn)
        self.__verify_pids()
        pid = self.dbus_info.GetConnectionUnixProcessID(sender)
        
        if self.authorisedPids.has_key(pid) and \
        self.authorisedPids[pid] == self.__get_start_time(pid):
            popen = subprocess.Popen(command, shell=intrShell, cwd=cwd,
                        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            outf, errf = popen.communicate()
            exitcode = popen.wait()
            return (dbus.UInt32(exitcode), dbus.String(outf), 
                dbus.String(errf))
        else:
            raise UnauthorisedPID("The process calling do() must call prepare() first")
    
    @dbus.service.method("org.tuxfamily.epidermis.PShellInterface",
                        in_signature='', out_signature='',
                        sender_keyword='sender', connection_keyword='conn')
    def leave(self, sender=None, conn=None):
        """Processes call this method to indicate that will no longer need
        to call do(). If they do, they can call prepare() again. This method
        may end the service if no other processes need it."""
        self.__setup_dbus_info(conn)
        pid = self.dbus_info.GetConnectionUnixProcessID(sender)
        
        if pid in self.authorisedPids:
            # remove pid from dictionary of authorised PIDs
            self.authorisedPids.pop(pid)
        self.__verify_pids()
    
    def __setup_dbus_info(self, conn):
        if self.dbus_info is None:
            self.dbus_info = dbus.Interface(conn.get_object('org.freedesktop.DBus',
                '/org/freedesktop/DBus/Bus', False), 'org.freedesktop.DBus')
    
    @staticmethod
    def __get_start_time(pid):
        """Returns the start-time of linux process (an integer relative
        to the kernel's start-time"""
        stat = open(os.path.join("/proc", str(pid), "stat"), "rb").read()
        time = stat.split(' ')[21]
        return int(time)
    
    @staticmethod
    def __verify_pid_is_epidermis(pid):
        """Checks whether the pid passed is Epidermis or not, if not,
        raises an exception"""
        cmdline = open(os.path.join("/proc",str(pid),"cmdline"),"rb").read().split(chr(0))
        if cmdline[-1] == '':
            cmdline = cmdline[:-1]
        if len(cmdline) < 2:
            logger.debug("TOO SHORT COMMANDLINE (sys.argv is %s)" % str(sys.argv))
            raise UnauthorisedPID("The process calling the DBus method is not Epidermis")
            
        # get current python executable
        python_exec = os.path.realpath(sys.executable)
        # strange bug: sometimes sys.executable returns the current directory
        # if it's a directory, find python using whereis on the command-line
        if os.path.isdir(python_exec):
            python_exec = None
            import subprocess
            pop = subprocess.Popen(["whereis","-b","python2.6"],stdout=subprocess.PIPE)
            output = pop.stdout.readlines()[0][:-1]
            for filename in output.split(" "):
               if filename[-10:] == "/python2.6" and "/bin/" in output and \
               filename[0] == "/":
                   python_exec = os.path.realpath(filename)
                   break
            logger.debug("Set python_exec to \"%s\" through whereis" % python_exec)
        if python_exec is None:
            class CantFindPythonExec(Exception):
                def __str__(self): return "Can't find python executable"
            raise CantFindPythonExec()
        # find environment variables of pid's process
        environ = open(os.path.join("/proc",str(pid),"environ"),
            "rb").read().split(chr(0))
        # set default path if not found:
        path = os.environ.get("PATH", "").split(':')
        # find path of process
        for ii in environ:
            if ii.split('=')[0] == "PATH":
                path = ii.split('=')[1].split(':')
        # get path of the first parameter in cmdline
        first = None
        for pp in path:
            if os.path.exists(os.path.join(pp, cmdline[0])):
                first = os.path.realpath(os.path.join(pp,cmdline[0]))
        if first is None:
            class ExecutableNotFound(Exception):
                def __str__(self): return "Executable not found of process"
            raise ExecutableNotFound()
        # check that the first parameter and the python executable are the
        # same
        try:
            if not os.path.samefile(first, python_exec):
                logger.debug("NOT PYTHON: %s != %s (sys.executable is %s)" % (first, python_exec, sys.executable))
                raise UnauthorisedPID("The process calling the DBus method is not Epidermis")
        except OSError, ee:
            if ee.errno == 2: # file not found
                logger.debug("FILE NOT FOUND (%s or %s)" % (first, python_exec))
                raise UnauthorisedPID("The process calling the DBus method is not Epidermis")
        
        # now check that the second parameter and Epidermis are the same
        cwd = os.path.realpath(os.path.join("/proc",str(pid),"cwd"))
        second = os.path.realpath(os.path.join(cwd, cmdline[1]))
        epidermis_exec = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), "epidermis.py")
        if not os.path.samefile(second, epidermis_exec):
            logger.debug("NOT EPIDERMIS, %s %s " % (second, epidermis_exec))
            raise UnauthorisedPID("The process calling the DBus method is not Epidermis")
        return True
        

    
    def __verify_pids(self):
        """Verify that the PIDs in self.authorisedPids still exist, and if
        no PIDs are left, end the service by calling self.mainloop.quit()"""
        for pid, start_time in self.authorisedPids.items():
            if not os.path.exists(os.path.join("/proc", str(pid), "stat")):
                self.authorisedPids.pop(pid)
            else:
                real_start_time = self.__get_start_time(pid)
                if start_time != real_start_time:
                    self.authorisedPids.pop(pid)
        if len(self.authorisedPids) == 0:
            print "self.mainloop.quit()"
            self.mainloop.quit()


def main():
    if len(sys.argv) >= 2 and "--service" in sys.argv[1:]:
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        mainloop = gobject.MainLoop()

        bus = dbus.SystemBus()
        name = dbus.service.BusName("org.tuxfamily.epidermis.Shell", bus)
        object = __PShell(mainloop, bus, '/org/tuxfamily/epidermis/PShell')

        print "Running epidermis/pshell_service.py service."
        mainloop.run()
    else:
        print >> sys.stderr, "Missing argument: --service"
        sys.exit(44)

if __name__ == "__main__":
    main()
